/*#define DEBUG*/
/*#define STANDALONE*/
/*
 *	gs-view.c
 *
 *		Revision 1.4  1992/09/16  04:42:23  hiro@jaist-east.ac.jp
 *		Revision 1.4a 1993/09/02  23:12:34  miyu@ftp.tohoku.ac.jp
 *
 * The source code is from xdvi "miyu" patchset, and obeys GPL.
 */

#include <stdio.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xmu/Atoms.h>

/*#include "xdvi.h"*/
#include "gs-view.h"

#include <fcntl.h>
#include <sys/stat.h>

#include <signal.h>
#include <setjmp.h>

extern int errno;
extern char *gsdevice;

#ifndef	SEEK_SET
#define	SEEK_SET	0
#endif

#ifndef	GHOSTSCRIPT
#define	GHOSTSCRIPT	"gs"
#endif

#define	DEV_NULL	"/dev/null"

struct gs_prop {
	Window	parent;
	Atom	GS_DONE;
	Atom	GS_PAGE;
};

static int
round(x)
	float x;
{
	if (x > 0)
	    return (int) (x + 0.5);
	else
	    return (int) (x - 0.5);
/*
	if (x > 0)
	    return (int) (x + ( 1 - (int)  0.5) / 2.0);
	else
	    return (int) (x + (-1 - (int) -0.5) / 2.0);
 */
}

static jmp_buf	abort_gs;

static RETSIGTYPE
catch_sig_child()
{
	/* longjmp(abort_gs, SIGCHLD); */
}

static RETSIGTYPE
catch_sig_user1()
{
	longjmp(abort_gs, SIGUSR1);
	/*NOTREACHED*/
}

static	Boolean
check_gs_event(disp, event, props)
	Display	*disp;
	XEvent	*event;
	struct gs_prop	*props;
{
	if (event->type               == ClientMessage  &&
	    event->xclient.window     == props-> parent &&
	    event->xclient.send_event == TRUE) {
#ifdef	DEBUG_XDVI_PS
		fprintf(stderr, "caught ClientMessage event\n");
#endif
	    if (event->xclient.message_type == props->GS_DONE ||
		event->xclient.message_type == props->GS_PAGE)
		    return TRUE;
	}
	else
	    return FALSE;
}

int
psfig_to_bitmap(gs)
	struct gs_info	*gs;
{
	char param[BUFSIZ];
	char devopt[BUFSIZ];
	RETSIGTYPE	(*old_sigchld)(), (*old_sigusr1)();

	sprintf(param, "0 0 %d %d %d %d %f %f 0 0 0 0 %d %d",
		round(gs->llx ), round(gs->lly ),
		round(gs->urx ), round(gs->ury ),
		gs->width  / ((gs->urx - gs->llx) / 72.0),
		gs->height / ((gs->ury - gs->lly) / 72.0),
		gs->width, gs->height);
	XChangeProperty(gs->disp, gs->parent,
			XInternAtom(gs->disp, "GHOSTVIEW", FALSE), XA_STRING,
			8, PropModeReplace, param, strlen(param));

#ifdef	DEBUG
	fprintf(stderr, "property:%s\n", param);
#endif

#ifdef	PS_LITERAL
	fflush(temp_ps_file);
	lseek(fileno(temp_ps_file), 0L, SEEK_SET);
#endif	/* PS_LITERAL */

	{
	    int val;

	    val = setjmp(abort_gs);
	    if (val == SIGUSR1) {
		fprintf(stderr, "GS terminated abnormally.\n");
		wait(NULL);
#ifdef	PS_LITERAL
		temp_ps_flush();
#endif	/* PS_LITERAL */
		return 1;
	    }
	}

	old_sigchld = signal(SIGCHLD, catch_sig_child);
	old_sigusr1 = signal(SIGUSR1, catch_sig_user1);

	if ((gs->gs_pid = fork()) == 0) {
	    /* child */
	    static char	env[BUFSIZ];
	    char	**argv;
	    int	i;
	    struct font	*fontp;
#ifdef	PS_HEADER
	    struct ps_hdr	*p;
#endif	/* PS_HEADER */

	    sprintf(env, "DISPLAY=%s", XDisplayString(gs->disp));
	    putenv(strdup(env));
	    sprintf(env, "GHOSTVIEW=%d %d", gs->parent, gs->pix);
	    putenv(strdup(env));

#ifdef	DEBUG
	    fprintf(stderr, "getenv:%s\n", getenv("DISPLAY"));
	    fprintf(stderr, "getenv:%s\n", getenv("GHOSTVIEW"));
#endif

	    close(fileno(stdin));
	    errno = 0;
#ifdef	PS_LITERAL
	    if (dup(fileno(temp_ps_file)) != fileno(stdin))
#else
	    if (open(DEV_NULL, O_RDONLY) != fileno(stdin))
#endif	/* PS_LITERAL */
	    {
		perror("Cound not redirect stdin");
		goto gs_exec_error;
	    }

	    close(fileno(stdout));
	    errno = 0;
	    if (open(DEV_NULL, O_WRONLY, 0777) != fileno(stdout)) {
		perror("Cound not redirect stdout");
		goto gs_exec_error;
	    }

#if 0
	    for (fontp = font_head; fontp != NULL; fontp = fontp->next)
		if (fontp->file != NULL) {
		    Fclose(fontp->file);
		    fontp->file = NULL;
		    ++n_files_left;
		}
#endif

	    i = 0;
#ifdef	PS_HEADER
	    for (p = ps_headers; p != NULL; p = p->next)
		i++;
#endif	/* PS_HEADER */
	    argv = (char **) malloc(sizeof(char *) * (i + 7));
	    if (argv == NULL) {
		fprintf(stderr, "could not exec gs\n");
		goto gs_exec_error;
	    }

	    i = 0;
	    argv[i++] = GHOSTSCRIPT;
	    argv[i++] = "-dQUIET";
	    argv[i++] = "-dNOPAUSE";
	    if (gsdevice) {
		sprintf(devopt, "-sDEVICE=%s", gsdevice);
		argv[i++] = devopt;
	    }
#ifdef	PS_HEADER
	    for (p = ps_headers; p != NULL; p = p->next)
		argv[i++] = p->psheader;
#endif	/* PS_HEADER */
#if 0
	    argv[i++] = "-";
#endif
	    argv[i++] = gs->psfile;
	    argv[i++] = "quit.ps";
	    argv[i++] = NULL;


#ifdef	DEBUG
	    fprintf(stderr, "gs:exec-ed\n");
	    for (i = 0; argv[i]; i++)
		fprintf(stderr, "arg[%d] = \"%s\"\n", i, argv[i]);
#endif
#ifdef	DEBUG
	    fprintf(stderr, "getenv:%s\n", getenv("DISPLAY"));
	    fprintf(stderr, "getenv:%s\n", getenv("GHOSTVIEW"));
#endif
	    execvp(GHOSTSCRIPT, argv);
	    perror("gs");

      gs_exec_error:
	    kill(getppid(), SIGUSR1);
	    exit(1);
	}
	else {
	    /* parent */
	    XEvent	ev;
	    struct gs_prop	props;

	    props.parent = gs->parent;
	    props.GS_DONE = XInternAtom(gs->disp, "DONE", False);
	    props.GS_PAGE = XInternAtom(gs->disp, "PAGE", False);

	    XIfEvent(gs->disp, &ev, check_gs_event, &props);
#ifdef	DEBUG
	    if (ev.xclient.message_type == props.GS_DONE ||
		ev.xclient.message_type == props.GS_PAGE) {
		fprintf(stderr, "caught event:GS_DONE or GS_PAGE\n");
		fprintf(stderr, "mess:%s\n", ev.xclient.data.b);
	    } else {
		fprintf(stderr, "caught another event\n");
	    }
#endif
	    signal(SIGCHLD, old_sigchld);
	    signal(SIGUSR1, old_sigusr1);
	    kill(gs-> gs_pid, SIGTERM);
	    wait(NULL);
	}

#ifdef	PS_LITERAL
	temp_ps_flush();
#endif	/* PS_LITERAL */

	return 0;
}

#ifdef	DEBUG
#ifdef STANDALONE
int
main(argc, argv)
	int	argc;
	char	**argv;
{
	struct gs_info	gs;
	GC	gc;
	int 	screen;

	gs.width = 400;
	gs.height = 400;

	gs.disp = XOpenDisplay("");
	screen = DefaultScreen(gs.disp);
	gs.parent = XCreateSimpleWindow(gs.disp, DefaultRootWindow(gs.disp),
					0, 0, gs.width, gs.height, 1,
					BlackPixel(gs.disp, screen),
					BlackPixel(gs.disp, screen));
	XMapWindow(gs.disp, gs.parent);

	gc = XCreateGC(gs.disp, gs.parent, 0, 0);
	XSetForeground(gs.disp, gc, WhitePixel(gs.disp, screen));
	XSetBackground(gs.disp, gc, BlackPixel(gs.disp, screen));

	XFlush(gs.disp);

	gs.psfile = argv[1];

    {
	int x1, y1, x2, y2;
	ps_boundingbox(gs.psfile, &x1, &y1, &x2, &y2);
	gs.llx = x1 + 0.0;
	gs.lly = y1 + 0.0;
	gs.urx = x2 + 0.0;
	gs.ury = y2 + 0.0;
    }

	fprintf(stderr, "bounding box: llx:%d lly:%d urx:%d ury:%d\n",
	       (int) gs.llx, (int) gs.lly, (int) gs.urx, (int) gs.ury);

	gs.pix = XCreatePixmap(gs.disp, gs.parent, gs.width, gs.height, 1);

	psfig_to_bitmap(&gs);

	XCopyArea(gs.disp, gs.pix, gs.parent, gc,
		  0, 0, 400, 400, 0, 0);
	XFlush(gs.disp);

	getchar();

	return 0;
}
#endif
#endif
